#ifndef MSGLIB
#define MSGLIB

#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>

typedef struct{
    int max_size; // maximum number of messages (can be extended, only used in reallocs)
    int size; // number of messages
    double * msgs; // messages
    int ndest; // number of destinations
    void ** dest; // destinations
} buffer;

void * initBuffer(){
    buffer* b = (buffer*)malloc(sizeof(buffer));
    b->ndest = 0;
    b->dest = (void**)malloc(5*sizeof(void*));
    b->max_size = 20;
    b->msgs = (double *)malloc(b->max_size*sizeof(double));
    b->size = 0;
    return (void*) b;

}

void closeBuffer(void* o){
    buffer* b = (buffer*) o;
    int i;
    free(b->msgs);
    free(b->dest);
    free(b);
}

// size of buffer
int bsize(void* o){
    buffer* b = (buffer*) o;
    return b->size;
}


void couple(void* o, void* d){
    buffer* b = (buffer*) o;

    b->ndest++;
    if (b->ndest > 5){ // initially allocated memory for five destinations in initBuffer
	b->dest = (void**)realloc(b->dest,b->ndest*sizeof(void*));	
    }
    b->dest[b->ndest-1] = d;
}

// put a new msg in the buffer o at position mindex
void bput(const double msg, void* o, int mindex){
    buffer* b = (buffer*) o;
    buffer* d;
    int i,j;
    int rindex;
    int index = mindex -1; //modelica first index == 1
    int rmax = 10;
    buffer** r = (buffer**)malloc(rmax*sizeof(buffer*)); // list of destination buffers that also have destinations (for coupled models).
    
    if (b->ndest == 0){ // no destinations, store msg locally	
	if (b->size == b->max_size){
	    b->max_size = b->max_size*2;
	    b->msgs = (double*)realloc(b->msgs,b->max_size*sizeof(double));
	}	
	b->size++;
	if (index < 0 || index == b->size-1){ // msg to the end of the buffer
	    b->msgs[b->size-1] = msg;
	}else{ // mst to a middle position
	    for(i=b->size-1;i>index;i--){
		b->msgs[i] = b->msgs[i-1];
		//	b->msgs[i] = (char*)realloc(b->msgs[i],(strlen(b->msgs[i-1])+1)*sizeof(char)); // strlen+1 to include terminating null ('\0') byte.
		//strcpy(b->msgs[i],b->msgs[i-1]);		
	    }
	    b->msgs[index] = msg;
	}
	//	b->msgs[b->size-1] = msg;
    }else{ // send msg to destinations
	rindex = 0;
	j = 0;
	r[0] = b;
	while(j <= rindex){ // for each buffer in the routing list
	    for (i = 0;i<r[j]->ndest;i++){ // for each destination in the current buffer
		d = (buffer*)r[j]->dest[i];
		if (d->ndest > 0){// include buffer in routing
		    rindex++;
		    if (rindex == rmax){ // realloc r if index reaches max
			rmax = rmax*2;
			r = (buffer**)realloc(r,rmax*sizeof(buffer*));
		    }
		    r[rindex] = d;
		}else{ // copy message in real destination (not routing)
		    if (d->size == d->max_size){
			d->max_size = d->max_size*2;
			d->msgs = (double*)realloc(d->msgs,d->max_size*sizeof(double));
		    }	
		    d->size++;
		    if (index < 0 || index == d->size-1){ // msg to the end of the buffer
			d->msgs[d->size-1] = msg;
		    }else{ // mst to a middle position
			for(i=d->size-1;i>index;i--){
			    d->msgs[i] = d->msgs[i-1];
			    //	b->msgs[i] = (char*)realloc(b->msgs[i],(strlen(b->msgs[i-1])+1)*sizeof(char)); // strlen+1 to include terminating null ('\0') byte.
			    //strcpy(b->msgs[i],b->msgs[i-1]);		
			}
			d->msgs[index] = msg;
		    }
		}
	    }
	    j++;
	}
    }
    free(r);
}

// get the indexed msg from the buffer, removing it
double bget(void* o, int mindex){
    buffer* b = (buffer*) o;
    double out;
    int i;
    int index = mindex -1;

    if (b->size >= index){
	out = b->msgs[index];
	for(i=index;i<b->size;i++){
	    b->msgs[i] = b->msgs[i+1];
	}
	b->size--;
	return out;
    }else
	return 0;
}

// read the indexed msg from the buffer
double bread(void* o, int mindex){
    buffer* b = (buffer*) o;
    double out;
    int i;
    int index = mindex -1;

    printf("BREAD o = %p, b = %p\n",o,b);

    printf("bread start\n");
    if (b->size >= index){
	out = b->msgs[index];
	//printf("read %s\n",out);
	//return out;
    }else{
	out = 0;
	//printf("read empty\n");
	//return (char*)"";    
    }
    printf("out = %lf\n",out);
    printf("END BREAD o = %p, b = %p\n",o,b);
    return out;
       
}


#endif










