/**********************************************************
* Author: 
*        Victorino Sanz 
*        Dpto. Informática y Automática, UNED
*        Juan del Rosal, 16
*        28040, Madrid
*        Spain
*        Email: vsanz@dia.uned.es
*
* Licensed by Victorino Sanz under the Modelica License 2
* Copyright © 2013, Victorino Sanz.
* 
***********************************************************/

#ifndef GNUPLOT
#define GNUPLOT

typedef struct ENV{
    int X;
    int Y;
    int nVars;
    int plot_var;
    double ***data;
    FILE **gp;
    char *path;
    int lastID;
    int animation;
} ENV;

void * CreateENV(int x, int y,int nVars,int plot_var,const char *path,const char *name,int plot_range,int plot_animation){
    ENV *e;
    int i,j,k;
    
    e = (ENV*)malloc(sizeof(ENV));
    e->X = x;
    e->Y = y;
    e->nVars = nVars;
    e->plot_var = plot_var-1;
    e->lastID = 0;
    e->animation = plot_animation;

    e->path = (char*)malloc(strlen(path)+1*sizeof(char));
    strcpy(e->path,path);
    
    e->data = (double***)malloc(x*sizeof(double**));
    for(i=0;i<x;i++)
	e->data[i] = (double**)malloc(y*sizeof(double*));
    for(i=0;i<x;i++)
	for(j=0;j<y;j++)
	    e->data[i][j] = (double*)malloc(nVars*sizeof(double));
    
    for(i=0;i<x;i++){
	for(j=0;j<y;j++){
	    for(k=0;k<nVars;k++){
		e->data[i][j][k] = 0;
	    }
	}
    }

    if (plot_animation) {
	    e->gp = (FILE**)malloc(nVars*sizeof(FILE*));
	    
#ifdef _WIN32
	    e->gp[0] = _popen(path, "w");
#else
	    e->gp[0] = popen(path, "w");
#endif
	    
	    if(e->gp[0] == NULL){
		fprintf(stderr, "Oops, I can't find %s.", path);
		exit(EXIT_FAILURE);
	    }    
#ifdef _WIN32
	    fprintf(e->gp[0], "set terminal wxt persist \n");
#else		  
	    fprintf(e->gp[0], "set terminal x11 \n");
#endif		  
	    fprintf(e->gp[0], "set title \"%s\" \n",name);
	    fprintf(e->gp[0], "set xrange [%d:0] \n",x);
	    fprintf(e->gp[0], "set yrange [%d:0] \n",y);
	    fprintf(e->gp[0], "set autoscale xy \n");
	    fprintf(e->gp[0], "set cbrange [0:%d] \n",plot_range);	    
	}
    
    return (void *)e;
}

int DeleteENV(void *environment){
    ENV *e;

    e = (ENV*)environment;

    free(e->data);
    return 1;
}

int Addplot(void *environment,int plot_var,int plot_range,const char *name){
    ENV *e;
    int var = plot_var-1;
    
    e = (ENV*)environment;

    if (e->animation){
#ifdef _WIN32
	e->gp[var] = _popen(e->path, "w");
#else
	e->gp[var] = popen(e->path, "w");
#endif
	
	if(e->gp[var] == NULL){
	    fprintf(stderr, "Oops, I can't find %s.", e->path);
	    exit(EXIT_FAILURE);
	}    
#ifdef _WIN32
	fprintf(e->gp[var], "set terminal wxt persist \n");
#else
	fprintf(e->gp[var], "set terminal x11 \n");
#endif
	fprintf(e->gp[var], "set title \"%s\" \n",name);
	fprintf(e->gp[var], "set xrange [0:%d] \n",e->X);
	fprintf(e->gp[var], "set yrange [0:%d] \n",e->Y);
	fprintf(e->gp[var], "set autoscale xy \n");
	fprintf(e->gp[var], "set cbrange [0:%d] \n",plot_range);
    }
    
    return 1;    
}


int newID(void *environment){
    ENV*e = (ENV*)environment;

    e->lastID++;
    return (e->lastID);
}

// Function that describes the state of initialized cells (non-initialized cells will remain in the default state)
int plot(void *environment,int plot_var){
    ENV*e = (ENV*)environment;
    int i,j;
    int var;

    //printf("PLOT\n");
    if (e->animation){
	if (plot_var <= 0)
	    var = 0;
	else
	    var = plot_var-1;
	
	// printf("PLOT\n");
	fprintf(e->gp[var],"plot \'-\' matrix with image\n");
	for(j=0;j<e->Y;j++){
	    for(i=0;i<e->X;i++){
		//	    printf("%f ",e->data[i][j][e->plot_var]);
		fprintf(e->gp[var],"%f ",e->data[i][j][var]);
	    }
	    fprintf(e->gp[var],"\n ");
	    //printf("\n ");
	}
	fprintf(e->gp[var],"\n");
	fprintf(e->gp[var],"e\n");
	fflush(e->gp[var]);	
    }
    
    return 1;
}

int clear(void *environment, int mvar){
    ENV*e = (ENV*)environment;
    int i,j;
    int var = mvar-1;

    
    for(i=0;i<e->X;i++){
	for(j=0;j<e->Y;j++){
	    e->data[i][j][var] = 0;
	}
    }
    return 1;
}

int modify(void *environment, int posx, int posy, int mvar, double value){
    ENV*e = (ENV*)environment;
    int x = posx-1;
    int y = posy-1;
    int var = mvar -1;

    //printf("UPDATE\n");
    e->data[x][y][var] += value;

    return 1;
}

int set(void *environment, int posx, int posy, int mvar, double value){
    ENV*e = (ENV*)environment;
    int x = posx-1;
    int y = posy-1;
    int var = mvar -1;

    //printf("SET\n");
    e->data[x][y][var] = value;
    return 1;
}

double readpos(void *environment, int posx, int posy, int mvar){
    ENV*e = (ENV*)environment;
    int x = posx-1;
    int y = posy-1;
    int var = mvar -1;
    //   printf("READ\n");
    return e->data[x][y][var];
}


#endif
