#ifndef MSGLIB
#define MSGLIB

#define LEN 10 // length of the arrays that represent each message
#define INITMAXSIZE 20 // initial maximum number of messages per buffer (can be reallocated) 
#define INITNDEST 5  // initial maximum number of destinations per buffer

#include <string.h>
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <ModelicaUtilities.h>
 


typedef struct{
    int max_size; // maximum number of messages (can be extended, only used in reallocs)
    int size; // number of messages
    char ** msgs; // messages
    int ndest; // number of destinations
    void ** dest; // destinations
} bufferS;

typedef struct{
    int max_size; // maximum number of messages (can be extended, only used in reallocs)
    int size; // number of messages
    double * msgs; // messages
    int ndest; // number of destinations
    void ** dest; // destinations
} bufferR;

typedef struct{
    int max_size; // maximum number of messages (can be extended, only used in reallocs)
    int size; // number of messages
    double ** msgs; // messages
    int *msglen; // length of each message in the buffer
    int ndest; // number of destinations
    void ** dest; // destinations
} bufferA;

/*********************************************************************************/

void * initBufferS(){
    bufferS* b = (bufferS*)malloc(sizeof(bufferS));
    b->ndest = 0;
    b->dest = (void**)malloc(INITNDEST*sizeof(void*));
    b->max_size = INITMAXSIZE;
    b->msgs = (char**)malloc(b->max_size*sizeof(char*));
    b->size = 0;
    return (void*) b;   
}

void * initBufferR(){
    bufferR* b = (bufferR*)malloc(sizeof(bufferR));
    b->ndest = 0;
    b->dest = (void**)malloc(INITNDEST*sizeof(void*));
    b->max_size = INITMAXSIZE;
    b->msgs = (double *)malloc(b->max_size*sizeof(double));
    b->size = 0;
    return (void*) b;
}

void * initBufferA(){
    bufferA* b = (bufferA*)malloc(sizeof(bufferA));
    b->ndest = 0;
    b->dest = (void**)malloc(INITNDEST*sizeof(void*));
    b->max_size = INITMAXSIZE;
    b->msgs = (double **)malloc(b->max_size*sizeof(double*));
    b->msglen = (int *)malloc(b->max_size*sizeof(int));
    b->size = 0;
    return (void*) b;
}

/*********************************************************************************/

void closeBufferS(void* o){
	bufferS* b = (bufferS*) o;
	free(b->msgs);
	free(b->dest);
	free(b);
}

void closeBufferR(void* o){
	bufferR* b = (bufferR*) o;
	//	free(b->msgs);
	//free(b->dest);
	//free(b);
}

void closeBufferA(void* o){
	bufferA* b = (bufferA*) o;
	//  free(b->msgs);
	//free(b->dest);
	free(b);
}

/*********************************************************************************/

// size of buffer
int bsizeS(void* o){
    bufferS* b = (bufferS*) o;
    return b->size;
}

int bsizeR(void* o){
    bufferR* b = (bufferR*) o;
    return b->size;
}

int bsizeA(void* o){
    bufferA* b = (bufferA*) o;
    return b->size;
}

/*********************************************************************************/

void coupleS(void* o, void* d){
    bufferS* b = (bufferS*) o;
    b->ndest++;
    if (b->ndest > INITNDEST){ // initially allocated memory for five destinations in initBuffer
	b->dest = (void**)realloc(b->dest,b->ndest*sizeof(void*));	
    }
    b->dest[b->ndest-1] = d;
}

void coupleR(void* o, void* d){
    bufferR* b = (bufferR*) o;
    b->ndest++;
    if (b->ndest > INITNDEST){ // initially allocated memory for five destinations in initBuffer
	b->dest = (void**)realloc(b->dest,b->ndest*sizeof(void*));	
    }
    b->dest[b->ndest-1] = d;
}

void coupleA(void* o, void* d){
    bufferA* b = (bufferA*) o;
    b->ndest++;
    if (b->ndest > INITNDEST){ // initially allocated memory for five destinations in initBuffer
	b->dest = (void**)realloc(b->dest,b->ndest*sizeof(void*));	
    }
    b->dest[b->ndest-1] = d;
}

/*********************************************************************************/

// put a new msg in the buffer o at a given position mindex
void bputS(const char* msg, void* o, int mindex){
    bufferS* b = (bufferS*) o;
    bufferS* d;
    int len;
    int i,j;
    int rindex;
    int index = mindex -1; //modelica first index == 1
    int rmax = 10;
    bufferS** r = (bufferS**)malloc(rmax*sizeof(bufferS*)); // list of destination buffers that also have destinations (for coupled models).

    //ModelicaFormatMessage("O: %p, bput start, mindex = %d, msg = %s \n",o,mindex,msg);
    //if (b->ndest == 0){
    // store msg locally
    if (b->size == b->max_size){
	//printf("realloc\n");
	b->max_size = b->max_size*2;
	b->msgs = (char**)realloc(b->msgs,b->max_size*sizeof(char*));
    }	
    b->size++;
    if (index < 0 || index == b->size-1){ // msg to the end of the buffer
	len = strlen(msg); 
	b->msgs[b->size-1] = (char*)malloc((len+1)*sizeof(char)); // len+1 to include terminating null ('\0') byte.
	strcpy(b->msgs[b->size-1],msg);
    }else{ // mst to a middle position
	for(i=b->size-1;i>index;i--){
	    b->msgs[i] = b->msgs[i-1];
	    //	b->msgs[i] = (char*)realloc(b->msgs[i],(strlen(b->msgs[i-1])+1)*sizeof(char)); // strlen+1 to include terminating null ('\0') byte.
	    //strcpy(b->msgs[i],b->msgs[i-1]);		
	}
	b->msgs[index] = (char*)malloc((len+1)*sizeof(char)); // len+1 to include terminating null ('\0') byte.
	strcpy(b->msgs[index],msg);
    }
    /* }/*else{ // send msg to destinations
    rindex = 0;
	j = 0;
	r[0] = b;
	while(j <= rindex){ // for each buffer in the routing list
	    for (i = 0;i<r[j]->ndest;i++){ // for each destination in the current buffer
		d = (bufferS*)r[j]->dest[i];
		if (d->ndest > 0){// include buffer in routing
		    rindex++;
		    if (rindex == rmax){ // realloc r if index reaches max
			rmax = rmax*2;
			r = (bufferS**)realloc(r,rmax*sizeof(bufferS*));
		    }
		    r[rindex] = d;
		}else{ // copy message in real destination (not routing)
		    if (d->size == d->max_size){
			d->max_size = d->max_size*2;
			d->msgs = (char**)realloc(d->msgs,d->max_size*sizeof(char*));
		    }	
		    d->size++;
		    if (index < 0 || index == d->size-1){ // msg to the end of the buffer
			len = strlen(msg); 
			d->msgs[d->size-1] = (char*)malloc((len+1)*sizeof(char)); // len+1 to include terminating null ('\0') byte.
			strcpy(d->msgs[d->size-1],msg);
		    }else{ // mst to a middle position
			for(i=d->size-1;i>index;i--){
			    d->msgs[i] = d->msgs[i-1];
			    //d->msgs[i] = (char*)realloc(d->msgs[i],(strlen(d->msgs[i-1])+1)*sizeof(char)); // strlen+1 to include terminating null ('\0') byte.
			    //strcpy(d->msgs[i],d->msgs[i-1]);		
			}
			d->msgs[index] = (char*)malloc((len+1)*sizeof(char)); // len+1 to include terminating null ('\0') byte.
			strcpy(d->msgs[index],msg);
		    }
		}
	    }
	    j++;
	}
    }*/
    free(r);
    //    for(i=0;i<b->size;i++)
    //	printf("msgs[%d]:_%s_:%d\n",i,b->msgs[i],strlen(b->msgs[i]));
    
    //ModelicaFormatMessage("bput end\n");
}

// put a new msg in the buffer o at position mindex
void bputR(const double msg, void* o, int mindex){
    bufferR* b = (bufferR*) o;
    bufferR* d;
    int i,j;
    int rindex;
    int index = mindex -1; //modelica first index == 1
    int rmax = 10;
    bufferR** r = (bufferR**)malloc(rmax*sizeof(bufferR*)); // list of destination buffers that also have destinations (for coupled models).

    //ModelicaFormatMessage("BPUTR msg %lf \n",msg);
    //if (b->ndest == 0){
    // store msg locally	
    if (b->size == b->max_size){
	b->max_size = b->max_size*2;
	b->msgs = (double*)realloc(b->msgs,b->max_size*sizeof(double));
    }	
    b->size++;
    if (index < 0 || index == b->size-1){ // msg to the end of the buffer
	b->msgs[b->size-1] = msg;
    }else{ // mst to a middle position
	for(i=b->size-1;i>index;i--){
	    b->msgs[i] = b->msgs[i-1];
	    //	b->msgs[i] = (char*)realloc(b->msgs[i],(strlen(b->msgs[i-1])+1)*sizeof(char)); // strlen+1 to include terminating null ('\0') byte.
	    //strcpy(b->msgs[i],b->msgs[i-1]);		
	}
	b->msgs[index] = msg;
    }
    //	b->msgs[b->size-1] = msg;
	/*}else{ // send msg to destinations
	rindex = 0;
	j = 0;
	r[0] = b;
	while(j <= rindex){ // for each buffer in the routing list
	    for (i = 0;i<r[j]->ndest;i++){ // for each destination in the current buffer
		d = (bufferR*)r[j]->dest[i];
		if (d->ndest > 0){// include buffer in routing
		    rindex++;
		    if (rindex == rmax){ // realloc r if index reaches max
			rmax = rmax*2;
			r = (bufferR**)realloc(r,rmax*sizeof(bufferR*));
		    }
		    r[rindex] = d;
		}else{ // copy message in real destination (not routing)
		    if (d->size == d->max_size){
			d->max_size = d->max_size*2;
			d->msgs = (double*)realloc(d->msgs,d->max_size*sizeof(double));
		    }	
		    d->size++;
		    if (index < 0 || index == d->size-1){ // msg to the end of the buffer
			d->msgs[d->size-1] = msg;
		    }else{ // mst to a middle position
			for(i=d->size-1;i>index;i--){
			    d->msgs[i] = d->msgs[i-1];
			    //	b->msgs[i] = (char*)realloc(b->msgs[i],(strlen(b->msgs[i-1])+1)*sizeof(char)); // strlen+1 to include terminating null ('\0') byte.
			    //strcpy(b->msgs[i],b->msgs[i-1]);		
			}
			d->msgs[index] = msg;
		    }
		}
	    }
	    j++;
	}
	}*/
    free(r);
}

// put a new msg in the buffer o at position mindex
void bputA(double *msg, size_t len, void* o, int mindex){
    bufferA* b = (bufferA*) o;
    bufferA* d;
    int i,j,k;
    int rindex;
    int index = mindex -1; //modelica first index == 1
    int rmax = 10;
    bufferA** r = (bufferA**)malloc(rmax*sizeof(bufferA*)); // list of destination buffers that also have destinations (for coupled models).
    
    // if (b->ndest == 0){
    // store msg locally	
    if (b->size == b->max_size){
	b->max_size = b->max_size*2;
	b->msgs = (double**)realloc(b->msgs,b->max_size*sizeof(double*));
	b->msglen = (int*)realloc(b->msglen,b->max_size*sizeof(int));
    }	
    b->size++;
    if (index < 0 || index == b->size-1){ // msg to the end of the buffer
	b->msglen[b->size-1] = len;
	b->msgs[b->size-1] = (double*)malloc(len*sizeof(double));
	for(i=0;i<len;i++)
	    b->msgs[b->size-1][i] = msg[i];
    }else{ // mst to a middle position
	for(i=b->size-1;i>index;i--){
	    b->msgs[i] = b->msgs[i-1];
	    //	b->msgs[i] = (char*)realloc(b->msgs[i],(strlen(b->msgs[i-1])+1)*sizeof(char)); // strlen+1 to include terminating null ('\0') byte.
	    //strcpy(b->msgs[i],b->msgs[i-1]);		
	}
	b->msglen[index] = len;
	b->msgs[index] = (double*)malloc(len*sizeof(double));
	for(k=0;k<len;k++)
	    b->msgs[index][k] = msg[k];
	//b->msgs[b->size-1] = msg; 
    }
    //	b->msgs[b->size-1] = msg;
    /*}/*else{ // send msg to destinations
	rindex = 0;
	j = 0;
	r[0] = b;
	while(j <= rindex){ // for each buffer in the routing list
	    for (i = 0;i<r[j]->ndest;i++){ // for each destination in the current buffer
		d = (bufferA*)r[j]->dest[i];
		if (d->ndest > 0){// include buffer in routing
		    rindex++;
		    if (rindex == rmax){ // realloc r if index reaches max
			rmax = rmax*2;
			r = (bufferA**)realloc(r,rmax*sizeof(bufferA*));
		    }
		    r[rindex] = d;
		}else{ // copy message in real destination (not routing)
		    if (d->size == d->max_size){
			d->max_size = d->max_size*2;
			d->msgs = (double**)realloc(d->msgs,d->max_size*sizeof(double*));
		    }	
		    d->size++;
		    if (index < 0 || index == d->size-1){ // msg to the end of the buffer
			d->msgs[d->size-1] = (double*)malloc(len*sizeof(double));
			for(k=0;k<len;k++)
			    d->msgs[d->size-1][k] = msg[k];
		    }else{ // mst to a middle position
			for(i=d->size-1;i>index;i--)
			    d->msgs[i] = d->msgs[i-1];
			d->msgs[index] = (double*)malloc(len*sizeof(double));
			for(k=0;k<len;k++)
			    d->msgs[index][k] = msg[k];
		    }
		}
	    }
	    j++;
	}
	}*/
	
    free(r);
}




/*************************************************************************************************/

// get the indexed msg from the buffer, removing it
char * bgetS(void* o, int mindex){
    bufferS* b = (bufferS*) o;
    char * out;
    int i;
    int index = mindex -1;

    if (b->size > index){
	out = b->msgs[index];
	for(i=index;i<b->size;i++){
	    b->msgs[i] = b->msgs[i+1];
	}
	b->size--;
	return out;
    }else
	return (char*)"";
}

// get the indexed msg from the buffer, removing it
double bgetR(void* o, int mindex){
    bufferR* b = (bufferR*) o;
    double out;
    int i;
    int index = mindex -1;

    if (b->size > index){
	out = b->msgs[index];
	for(i=index;i<b->size;i++){
	    b->msgs[i] = b->msgs[i+1];
	}
	b->size--;
	return out;
    }else
	return 0;
}

// get the position mpos of indexed msg from the buffer, removing the whole msg
double bgetA(void* o, int mpos, int mindex){
    bufferA* b = (bufferA*) o;
    double * out;
    int i;
    int index = mindex -1;
    int pos = mpos -1;
    
    if (b->size > index){
	out = b->msgs[index];
	for(i=index;i<b->size;i++){
	    b->msgs[i] = b->msgs[i+1];
	}
	b->size--;
	return out[pos];	
    }
}


/******************************************************************************************************/

// read the indexed msg from the buffer
char * breadS(void* o, int mindex){
    bufferS* b = (bufferS*) o;
    char * out;
    int i;
    int index = mindex -1;

    if (b->size >= index){
	out = b->msgs[index];
    }else{
	out = (char*)"";
    }
    return out;
       
}

// read the indexed msg from the buffer
double breadR(void* o, int mindex){
    bufferR* b = (bufferR*) o;
    double out;
    int i;
    int index = mindex -1;

    if (b->size >= index){
	out = b->msgs[index];
    }else{
	out = 0;
    }
    return out;
       
}

// read the indexed msg from the buffer
double breadA(void* o, int mpos, int mindex){
    bufferA* b = (bufferA*) o;
    double * out;
    int i;
    int index = mindex -1;
    int pos = mpos -1;

    if (b->size >= index){
	out = b->msgs[index];
	return out[pos];
    }       
}


/*********************************************************************************/

// sends msgs stored in the buffer to its destinations
void bsendS(void* o){
    bufferS* b = (bufferS*) o;
    bufferS* d;
    int len;
    int i,j,k;
    int rindex;
    int bsize; // size of b
    //    int index = mindex -1; //modelica first index == 1
    int rmax = 10;
    bufferS** r = (bufferS**)malloc(rmax*sizeof(bufferS*)); // list of destination buffers that also have destinations (for coupled models).
    char **msgs;
    int *msgsize;
    
    //ModelicaFormatMessage("O: %p, bput start, mindex = %d, msg = %s \n",o,mindex,msg);
    bsize = bsizeS(o);
    msgs = (char **)malloc(bsize*sizeof(char*));
    msgsize = (int*)malloc(bsize*sizeof(int));
    for (i=0;i<bsize;i++){
	msgsize[i] = sizeof(b->msgs[i])/sizeof(char);
	msgs[i] = (char*)malloc(msgsize[i]*sizeof(char));
	msgs[i] = bgetS(o,1);
    }
	
    if (b->ndest > 0){ // send msg to destinations
	rindex = 0;
	j = 0;
	r[0] = b; // routing list
	while(j <= rindex){ // for each buffer in the routing list
	    for (i = 0;i<r[j]->ndest;i++){ // for each destination in the current buffer
		d = (bufferS*)r[j]->dest[i];
		if (d->ndest > 0){// include buffer in routing
		    rindex++;
		    if (rindex == rmax){ // realloc r if index reaches max
			rmax = rmax*2;
			r = (bufferS**)realloc(r,rmax*sizeof(bufferS*));
		    }
		    r[rindex] = d;
		}else // put messages into destination
		    for (k = 0;k<bsize;k++)
			bputS(msgs[k],d,0);		
	    }
	}
	j++;
    }
    free(r);
    //    for(i=0;i<b->size;i++)
    //	printf("msgs[%d]:_%s_:%d\n",i,b->msgs[i],strlen(b->msgs[i]));
    
    //ModelicaFormatMessage("bput end\n");
}

// put a new msg in the buffer o at position mindex
void bsendR(void* o){
    bufferR* b = (bufferR*) o;
    bufferR* d;
    int i,j,k;
    int rindex;
    //    int index = mindex -1; //modelica first index == 1
    int rmax = 10;
    bufferR** r = (bufferR**)malloc(rmax*sizeof(bufferR*)); // list of destination buffers that also have destinations (for coupled models).
    double *msgs;
    int bsize;

    //ModelicaFormatMessage("BPUTR msg %lf \n",msg);
    bsize = bsizeR(o);
    msgs = (double *)malloc(bsize*sizeof(double));
    for (i=0;i<bsize;i++)
	msgs[i] = bgetR(o,1);
    if (b->ndest > 0){// send msg to destinations
	rindex = 0;
	j = 0;
	r[0] = b;
	while(j <= rindex){ // for each buffer in the routing list
	    for (i = 0;i<r[j]->ndest;i++){ // for each destination in the current buffer
		d = (bufferR*)r[j]->dest[i];
		if (d->ndest > 0){// include buffer in routing
		    rindex++;
		    if (rindex == rmax){ // realloc r if index reaches max
			rmax = rmax*2;
			r = (bufferR**)realloc(r,rmax*sizeof(bufferR*));
		    }
		    r[rindex] = d;
		}else // copy message in real destination (not routing)
		    for (k = 0;k<bsize;k++)
			bputR(msgs[k],d,0);		
	    }
	    j++;
	}
    }
    free(r);
}

// put a new msg in the buffer o at position mindex
void bsendA(void* o){
    bufferA* b = (bufferA*) o;
    bufferA* d;
    int i,j,k;
    int rindex;
    //    int index = mindex -1; //modelica first index == 1
    int rmax = 10;
    bufferA** r = (bufferA**)malloc(rmax*sizeof(bufferA*)); // list of destination buffers that also have destinations (for coupled models).
    double **msgs;
    int *msgsize;
    int bsize;

    bsize = bsizeA(o);
    msgs = (double **)malloc(bsize*sizeof(double*));
    msgsize = (int *)malloc(bsize*sizeof(int));
    for (i=0;i<bsize;i++){
	msgsize[i] = b->msglen[i];
	msgs[i] = (double*)malloc(msgsize[i]*sizeof(double));
	for(j=0;j<msgsize[i]-1;j++){
	    msgs[i][j] = breadA(o,j+1,1); // read positions of the array
	}
	msgs[i][j] = bgetA(o,j+1,1); // get last position and remove msg from buffer
    }

    if (b->ndest > 0){ // send msg to destinations
	rindex = 0;
	j = 0;
	r[0] = b;
	while(j <= rindex){ // for each buffer in the routing list
	    for (i = 0;i<r[j]->ndest;i++){ // for each destination in the current buffer
		d = (bufferA*)r[j]->dest[i];
		if (d->ndest > 0){// include buffer in routing
		    rindex++;
		    if (rindex == rmax){ // realloc r if index reaches max
			rmax = rmax*2;
			r = (bufferA**)realloc(r,rmax*sizeof(bufferA*));
		    }
		    r[rindex] = d;
		}else{ // copy message in real destination (not routing)
		    for (k = 0;k<bsize;k++)
			bputA(msgs[k],msgsize[k],d,0);		
		}
	    }
	    j++;
	}
    }
    free(r);
}






#endif










