#ifndef MSGLIB
#define MSGLIB

#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>

typedef struct packet{
    int addro;
    int addrd;
    int packetsize;
    int role;
    double payload1;
    double payload2;
} packet;

typedef struct{
    int max_size; // maximum number of messages (can be extended, only used in reallocs)
    int size; // number of messages
    packet * msgs; // messages
    int ndest; // number of destinations
    void ** dest; // destinations
} buffer;

void * initBuffer(){
    buffer* b = (buffer*)malloc(sizeof(buffer));
    b->ndest = 0;
    b->dest = (void**)malloc(5*sizeof(void*));
    b->max_size = 20;
    b->msgs = (packet *)malloc(b->max_size*sizeof(packet));
    b->size = 0;
    return (void*) b;

}

void closeBuffer(void* o){
    buffer* b = (buffer*) o;
    int i;
    free(b->msgs);
    free(b->dest);
    free(b);
}

// size of buffer
int bsize(void* o){
    buffer* b = (buffer*) o;
    return b->size;
}


void couple(void* o, void* d){
    buffer* b = (buffer*) o;

    b->ndest++;
    if (b->ndest > 5){ // initially allocated memory for five destinations in initBuffer
	b->dest = (void**)realloc(b->dest,b->ndest*sizeof(void*));	
    }
    b->dest[b->ndest-1] = d;
}

// put a new msg at the end of a buffer
void bput(packet* msg, void* o){
    buffer* b = (buffer*) o;
    buffer* d;
    int i,j;
    int rindex;
    int rmax = 10;
    buffer** r = (buffer**)malloc(rmax*sizeof(buffer*)); // list of destination buffers that also have destinations (for coupled models).
    
    if (b->ndest == 0){ // no destinations, store msg locally	
	if (b->size == b->max_size){
	    b->max_size = b->max_size*2;
	    b->msgs = (packet*)realloc(b->msgs,b->max_size*sizeof(packet));
	}	

	b->size++;
	//printf("msg.addro = %d, msg.payload1 = %lf\n",msg->addro,msg->payload1);
	//b->msgs[b->size-1] = *msg;
	b->msgs[b->size-1].addro = msg->addro;
	b->msgs[b->size-1].addrd = msg->addrd;
	b->msgs[b->size-1].packetsize = msg->packetsize;
	b->msgs[b->size-1].role = msg->role;
	b->msgs[b->size-1].payload1 = msg->payload1;
	b->msgs[b->size-1].payload2 = msg->payload2;
	//printf("msgs.addro = %d, msgs.payload1 = %lf\n",b->msgs[b->size-1].addro,b->msgs[b->size-1].payload1);
    }else{ // send msg to destinations
	rindex = 0;
	j = 0;
	r[0] = b;
	while(j <= rindex){ // for each buffer in the routing list
	    for (i = 0;i<r[j]->ndest;i++){ // for each destination in the current buffer
		d = (buffer*)r[j]->dest[i];
		if (d->ndest > 0){// include buffer in routing
		    rindex++;
		    if (rindex == rmax){ // realloc r if index reaches max
			rmax = rmax*2;
			r = (buffer**)realloc(r,rmax*sizeof(buffer*));
		    }
		    r[rindex] = d;
		}else{ // copy message in real destination (not routing)
		    if (d->size == d->max_size){
			d->max_size = d->max_size*2;
			d->msgs = (packet*)realloc(d->msgs,d->max_size*sizeof(packet));
		    }	
		    d->size++;
		    //printf("Dmsg.addro = %d, Dmsg.payload1 = %lf\n",msg->addro,msg->payload1);
		    //d->msgs[d->size-1] = *msg;
		    d->msgs[d->size-1].addro = msg->addro;
		    d->msgs[d->size-1].addrd = msg->addrd;
		    d->msgs[d->size-1].packetsize = msg->packetsize;
		    d->msgs[d->size-1].role = msg->role;
		    d->msgs[d->size-1].payload1 = msg->payload1;
		    d->msgs[d->size-1].payload2 = msg->payload2;
		    //printf("msgs.addro = %d, msgs.payload1 = %lf\n",d->msgs[d->size-1].addro,d->msgs[d->size-1].payload1);
		}
	    }
	    j++;
	}
    }
    free(r);
}

// get the indexed msg from the buffer, removing it
void bget(void* o, int mindex, packet *out){
    buffer* b = (buffer*) o;
    int i;
    int index = mindex -1;

    out->addro = -1;
    out->addrd = -1;
    out->packetsize = 100;
    out->role = -5;
    out->payload1 = -1;
    out->payload2 = -1;
    if (b->size >= index){
	out->addro = b->msgs[index].addro;
	out->addrd = b->msgs[index].addrd;
	out->packetsize = b->msgs[index].packetsize;
	out->role = b->msgs[index].role;
	out->payload1 = b->msgs[index].payload1;
	out->payload2 = b->msgs[index].payload2;
	
	//	out = &(b->msgs[index]);
	for(i=index;i<b->size;i++){
	    b->msgs[i] = b->msgs[i+1];
	}
	b->size--;
    }   
}

#endif










